"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.createFileEmitter = exports.angular = void 0;
const core_1 = require("@babel/core");
const path = require("path");
const ts = require("typescript");
const vite_1 = require("vite");
const angular_build_optimizer_plugin_1 = require("./angular-build-optimizer-plugin");
const angular_jit_plugin_1 = require("./angular-jit-plugin");
const angular_vitest_plugin_1 = require("./angular-vitest-plugin");
const compiler_plugin_1 = require("./compiler-plugin");
const component_resolvers_1 = require("./component-resolvers");
const host_1 = require("./host");
const devkit_1 = require("./utils/devkit");
/**
 * TypeScript file extension regex
 * Match .(c or m)ts, .ts extensions with an optional ? for query params
 * Ignore .tsx extensions
 */
const TS_EXT_REGEX = /\.[cm]?(ts|analog)[^x]?\??/;
function angular(options) {
    /**
     * Normalize plugin options so defaults
     * are used for values not provided.
     */
    const pluginOptions = {
        tsconfig: options?.tsconfig ??
            (process.env['NODE_ENV'] === 'test'
                ? './tsconfig.spec.json'
                : './tsconfig.app.json'),
        workspaceRoot: options?.workspaceRoot ?? process.cwd(),
        inlineStylesExtension: options?.inlineStylesExtension ?? 'css',
        advanced: {
            tsTransformers: {
                before: options?.advanced?.tsTransformers?.before ?? [],
                after: options?.advanced?.tsTransformers?.after ?? [],
                afterDeclarations: options?.advanced?.tsTransformers?.afterDeclarations ?? [],
            },
        },
        supportedBrowsers: options?.supportedBrowsers ?? ['safari 15'],
        jit: options?.jit,
        supportAnalogFormat: options?.experimental?.supportAnalogFormat ?? false,
    };
    // The file emitter created during `onStart` that will be used during the build in `onLoad` callbacks for TS files
    let fileEmitter;
    let compilerOptions = {};
    // Temporary deep import for transformer support
    const { mergeTransformers, replaceBootstrap, } = require('@ngtools/webpack/src/ivy/transformation');
    const { augmentProgramWithVersioning, augmentHostWithCaching, } = require('@ngtools/webpack/src/ivy/host');
    let compilerCli;
    let userConfig;
    let rootNames;
    let host;
    let nextProgram;
    let builderProgram;
    let watchMode = false;
    const sourceFileCache = new devkit_1.SourceFileCache();
    const isProd = process.env['NODE_ENV'] === 'production';
    const isTest = process.env['NODE_ENV'] === 'test' || !!process.env['VITEST'];
    const jit = typeof pluginOptions?.jit !== 'undefined' ? pluginOptions.jit : isTest;
    let viteServer;
    let cssPlugin;
    let styleTransform;
    const styleUrlsResolver = new component_resolvers_1.StyleUrlsResolver();
    const templateUrlsResolver = new component_resolvers_1.TemplateUrlsResolver();
    function angularPlugin() {
        return {
            name: '@analogjs/vite-plugin-angular',
            async config(config, { command }) {
                watchMode = command === 'serve';
                userConfig = config;
                pluginOptions.tsconfig =
                    options?.tsconfig ??
                        path.resolve(config.root || '.', process.env['NODE_ENV'] === 'test'
                            ? './tsconfig.spec.json'
                            : './tsconfig.app.json');
                compilerCli = await (0, devkit_1.loadEsmModule)('@angular/compiler-cli');
                return {
                    optimizeDeps: {
                        include: ['rxjs/operators', 'rxjs'],
                        exclude: ['@angular/platform-server'],
                        esbuildOptions: {
                            plugins: [
                                (0, compiler_plugin_1.createCompilerPlugin)({
                                    tsconfig: pluginOptions.tsconfig,
                                    sourcemap: !isProd,
                                    advancedOptimizations: isProd,
                                    jit,
                                    incremental: watchMode,
                                }, isTest),
                            ],
                            define: {
                                ngJitMode: 'false',
                                ngI18nClosureMode: 'false',
                                ...(watchMode ? {} : { ngDevMode: 'false' }),
                            },
                        },
                    },
                    resolve: {
                        conditions: ['style'],
                    },
                };
            },
            configureServer(server) {
                viteServer = server;
                server.watcher.on('add', setupCompilation);
                server.watcher.on('unlink', setupCompilation);
            },
            async buildStart({ plugins }) {
                if (Array.isArray(plugins)) {
                    cssPlugin = plugins.find((plugin) => plugin.name === 'vite:css');
                }
                setupCompilation(userConfig);
                // Only store cache if in watch mode
                if (watchMode) {
                    augmentHostWithCaching(host, sourceFileCache);
                }
                await buildAndAnalyze();
            },
            async handleHotUpdate(ctx) {
                // The `handleHotUpdate` hook may be called before the `buildStart`,
                // which sets the compilation. As a result, the `host` may not be available
                // yet for use, leading to build errors such as "cannot read properties of undefined"
                // (because `host` is undefined).
                if (!host) {
                    return;
                }
                if (TS_EXT_REGEX.test(ctx.file)) {
                    sourceFileCache.invalidate([ctx.file.replace(/\?(.*)/, '')]);
                    await buildAndAnalyze();
                }
                if (/\.(html|htm|css|less|sass|scss)$/.test(ctx.file)) {
                    /**
                     * Check to see if this was a direct request
                     * for an external resource (styles, html).
                     */
                    const isDirect = ctx.modules.find((mod) => ctx.file === mod.file && mod.id?.includes('?direct'));
                    if (isDirect) {
                        return ctx.modules;
                    }
                    const mods = [];
                    ctx.modules.forEach((mod) => {
                        mod.importers.forEach((imp) => {
                            sourceFileCache.invalidate([imp.id]);
                            ctx.server.moduleGraph.invalidateModule(imp);
                            mods.push(imp);
                        });
                    });
                    await buildAndAnalyze();
                    return mods;
                }
                return ctx.modules;
            },
            async transform(code, id) {
                // Skip transforming node_modules
                if (id.includes('node_modules')) {
                    return;
                }
                /**
                 * Check for options.transformFilter
                 */
                if (options?.transformFilter &&
                    !(options?.transformFilter(code, id) ?? true)) {
                    return;
                }
                /**
                 * Check for .ts extenstions for inline script files being
                 * transformed (Astro).
                 *
                 * Example ID:
                 *
                 * /src/pages/index.astro?astro&type=script&index=0&lang.ts
                 */
                if (id.includes('type=script')) {
                    return;
                }
                if (TS_EXT_REGEX.test(id)) {
                    if (id.includes('.ts?')) {
                        // Strip the query string off the ID
                        // in case of a dynamically loaded file
                        id = id.replace(/\?(.*)/, '');
                    }
                    /**
                     * Re-analyze on each transform
                     * for test(Vitest)
                     */
                    if (isTest) {
                        const tsMod = viteServer?.moduleGraph.getModuleById(id);
                        if (tsMod) {
                            sourceFileCache.invalidate([id]);
                            await buildAndAnalyze();
                        }
                    }
                    let templateUrls = [];
                    let styleUrls = [];
                    if ((0, component_resolvers_1.hasTemplateUrl)(code)) {
                        templateUrls = templateUrlsResolver.resolve(code, id);
                    }
                    if ((0, component_resolvers_1.hasStyleUrls)(code)) {
                        styleUrls = styleUrlsResolver.resolve(code, id);
                    }
                    if (watchMode) {
                        for (const urlSet of [...templateUrls, ...styleUrls]) {
                            // `urlSet` is a string where a relative path is joined with an
                            // absolute path using the `|` symbol.
                            // For example: `./app.component.html|/home/projects/analog/src/app/app.component.html`.
                            const [, absoluteFileUrl] = urlSet.split('|');
                            this.addWatchFile(absoluteFileUrl);
                        }
                    }
                    const typescriptResult = fileEmitter && (await fileEmitter(id));
                    // return fileEmitter
                    let data = typescriptResult?.content ?? '';
                    if (jit && data.includes('angular:jit:')) {
                        data = data.replace(/angular:jit:style:inline;/g, 'virtual:angular:jit:style:inline;');
                        templateUrls.forEach((templateUrlSet) => {
                            const [templateFile, resolvedTemplateUrl] = templateUrlSet.split('|');
                            data = data.replace(`angular:jit:template:file;${templateFile}`, `${resolvedTemplateUrl}?raw`);
                        });
                        styleUrls.forEach((styleUrlSet) => {
                            const [styleFile, resolvedStyleUrl] = styleUrlSet.split('|');
                            data = data.replace(`angular:jit:style:file;${styleFile}`, `${resolvedStyleUrl}?inline`);
                        });
                    }
                    if (jit) {
                        return {
                            code: data.replace(/^\/\/# sourceMappingURL=[^\r\n]*/gm, ''),
                            map: {
                                mappings: '',
                            },
                        };
                    }
                    const forceAsyncTransformation = /for\s+await\s*\(|async\s+function\s*\*/.test(data);
                    const useInputSourcemap = (!isProd ? undefined : false);
                    if (id.includes('.analog') &&
                        pluginOptions.supportAnalogFormat &&
                        fileEmitter) {
                        sourceFileCache.invalidate([`${id}.ts`]);
                        const ngFileResult = await fileEmitter(`${id}.ts`);
                        data = ngFileResult?.content || '';
                    }
                    if (!forceAsyncTransformation && !isProd) {
                        return {
                            code: data.replace(/^\/\/# sourceMappingURL=[^\r\n]*/gm, ''),
                            map: {
                                mappings: '',
                            },
                        };
                    }
                    const babelResult = await (0, core_1.transformAsync)(data, {
                        filename: id,
                        inputSourceMap: (useInputSourcemap
                            ? undefined
                            : false),
                        sourceMaps: !isProd ? 'inline' : false,
                        compact: false,
                        configFile: false,
                        babelrc: false,
                        browserslistConfigFile: false,
                        plugins: [],
                        presets: [
                            [
                                devkit_1.angularApplicationPreset,
                                {
                                    supportedBrowsers: pluginOptions.supportedBrowsers,
                                    forceAsyncTransformation,
                                    optimize: isProd && {},
                                },
                            ],
                        ],
                    });
                    return {
                        code: babelResult?.code ?? '',
                        map: babelResult?.map,
                    };
                }
                return undefined;
            },
        };
    }
    return [
        angularPlugin(),
        (isTest && (0, angular_vitest_plugin_1.angularVitestPlugin)()),
        (jit &&
            (0, angular_jit_plugin_1.jitPlugin)({
                inlineStylesExtension: pluginOptions.inlineStylesExtension,
            })),
        (0, angular_build_optimizer_plugin_1.buildOptimizerPlugin)({
            isProd,
            supportedBrowsers: pluginOptions.supportedBrowsers,
        }),
    ].filter(Boolean);
    function findAnalogFiles(config) {
        if (!pluginOptions.supportAnalogFormat) {
            return [];
        }
        const fg = require('fast-glob');
        const root = (0, vite_1.normalizePath)(path.resolve(pluginOptions.workspaceRoot, config.root || '.'));
        return fg
            .sync([`${root}/**/*.analog`], {
            dot: true,
        })
            .map((file) => `${file}.ts`);
    }
    function setupCompilation(config) {
        const analogFiles = findAnalogFiles(config);
        const { options: tsCompilerOptions, rootNames: rn } = compilerCli.readConfiguration(pluginOptions.tsconfig, {
            suppressOutputPathCheck: true,
            outDir: undefined,
            inlineSourceMap: !isProd,
            inlineSources: !isProd,
            declaration: false,
            declarationMap: false,
            allowEmptyCodegenFiles: false,
            annotationsAs: 'decorators',
            enableResourceInlining: false,
            supportTestBed: false,
        });
        if (pluginOptions.supportAnalogFormat) {
            // Experimental Local Compilation is necessary
            // for the Angular compiler to work with
            // AOT and virtually compiled .analog files.
            tsCompilerOptions.compilationMode = 'experimental-local';
        }
        rootNames = rn.concat(analogFiles);
        compilerOptions = tsCompilerOptions;
        host = ts.createIncrementalCompilerHost(compilerOptions);
        styleTransform = watchMode
            ? viteServer.pluginContainer.transform
            : cssPlugin.transform;
        if (!jit) {
            (0, host_1.augmentHostWithResources)(host, styleTransform, {
                inlineStylesExtension: pluginOptions.inlineStylesExtension,
                supportAnalogFormat: pluginOptions.supportAnalogFormat,
                isProd,
            });
        }
    }
    /**
     * Creates a new NgtscProgram to analyze/re-analyze
     * the source files and create a file emitter.
     * This is shared between an initial build and a hot update.
     */
    async function buildAndAnalyze() {
        let builder;
        let typeScriptProgram;
        let angularCompiler;
        if (!jit) {
            // Create the Angular specific program that contains the Angular compiler
            const angularProgram = new compilerCli.NgtscProgram(rootNames, compilerOptions, host, nextProgram);
            angularCompiler = angularProgram.compiler;
            typeScriptProgram = angularProgram.getTsProgram();
            augmentProgramWithVersioning(typeScriptProgram);
            builder = builderProgram =
                ts.createEmitAndSemanticDiagnosticsBuilderProgram(typeScriptProgram, host, builderProgram);
            await angularCompiler.analyzeAsync();
            nextProgram = angularProgram;
        }
        else {
            builder = builderProgram =
                ts.createEmitAndSemanticDiagnosticsBuilderProgram(rootNames, compilerOptions, host, nextProgram);
            typeScriptProgram = builder.getProgram();
            nextProgram = builderProgram;
        }
        if (!watchMode) {
            // When not in watch mode, the startup cost of the incremental analysis can be avoided by
            // using an abstract builder that only wraps a TypeScript program.
            builder = ts.createAbstractBuilder(typeScriptProgram, host);
        }
        const getTypeChecker = () => builder.getProgram().getTypeChecker();
        fileEmitter = createFileEmitter(builder, mergeTransformers({
            before: [
                replaceBootstrap(getTypeChecker),
                ...(jit
                    ? [
                        compilerCli.constructorParametersDownlevelTransform(builder.getProgram()),
                        (0, devkit_1.createJitResourceTransformer)(getTypeChecker),
                    ]
                    : []),
                ...pluginOptions.advanced.tsTransformers.before,
            ],
            after: pluginOptions.advanced.tsTransformers.after,
            afterDeclarations: pluginOptions.advanced.tsTransformers.afterDeclarations,
        }, jit ? {} : angularCompiler.prepareEmit().transformers), () => []);
    }
}
exports.angular = angular;
function createFileEmitter(program, transformers = {}, onAfterEmit) {
    return async (file) => {
        const sourceFile = program.getSourceFile(file);
        if (!sourceFile) {
            return undefined;
        }
        let content;
        program.emit(sourceFile, (filename, data) => {
            if (/\.[cm]?js$/.test(filename)) {
                content = data;
            }
        }, undefined /* cancellationToken */, undefined /* emitOnlyDtsFiles */, transformers);
        onAfterEmit?.(sourceFile);
        return { content, dependencies: [] };
    };
}
exports.createFileEmitter = createFileEmitter;
//# sourceMappingURL=angular-vite-plugin.js.map